@@ -40,7 +40,41 @@ static int ItclCreateMemberCode(Tcl_Interp* interp, ItclClass *iclsPtr,
 static int ItclCreateMemberFunc(Tcl_Interp* interp, ItclClass *iclsPtr,
 	Tcl_Obj *namePtr, const char* arglist, const char* body,
         ItclMemberFunc** imPtrPtr, int flags);
+void    ItclFreeMemberCode (ItclMemberCode *mcodePtr);
 
+void
+ItclPreserveIMF(
+    ItclMemberFunc *imPtr)
+{
+    imPtr->refCount++;
+}
+
+void
+ItclReleaseIMF(
+    ClientData clientData)
+{
+    ItclMemberFunc *imPtr = (ItclMemberFunc *)clientData;
+
+    if (--imPtr->refCount == 0) {
+	Itcl_DeleteMemberFunc(clientData);
+    }
+}
+
+void
+ItclPreserveMemberCode(
+    ItclMemberCode *mcodePtr)
+{
+    mcodePtr->refCount++;
+}
+
+void
+ItclReleaseMemberCode(
+    ItclMemberCode *mcodePtr)
+{
+    if (--mcodePtr->refCount == 0) {
+	ItclFreeMemberCode(mcodePtr);
+    }
+}
 
 /*
  * ------------------------------------------------------------------------
@@ -276,11 +310,10 @@ NRConfigBodyCmd(
         goto configBodyCmdDone;
     }
 
-    Itcl_PreserveData((ClientData)mcode);
-    Itcl_EventuallyFree((ClientData)mcode, Itcl_DeleteMemberCode);
+    ItclPreserveMemberCode(mcode);
 
     if (ivPtr->codePtr) {
-        Itcl_ReleaseData((ClientData)ivPtr->codePtr);
+        ItclReleaseMemberCode(ivPtr->codePtr);
     }
     ivPtr->codePtr = mcode;
 
@@ -358,6 +391,7 @@ ItclCreateMethod(
         Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
             "bad method name \"", Tcl_GetString(namePtr), "\"",
             (char*)NULL);
+	Tcl_DecrRefCount(namePtr);
         return TCL_ERROR;
     }
 
@@ -462,7 +496,7 @@ ItclCreateMemberFunc(
      *  same name doesn't already exist.
      */
     hPtr = Tcl_CreateHashEntry(&iclsPtr->functions, (char *)namePtr, &newEntry);
-    if (!hPtr) {
+    if (!newEntry) {
         Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
             "\"", Tcl_GetString(namePtr), "\" already defined in class \"",
             Tcl_GetString(iclsPtr->fullNamePtr), "\"",
@@ -480,9 +514,6 @@ ItclCreateMemberFunc(
         return TCL_ERROR;
     }
 
-    Itcl_PreserveData((ClientData)mcode);
-    Itcl_EventuallyFree((ClientData)mcode, Itcl_DeleteMemberCode);
-
     /*
      *  Allocate a member function definition and return.
      */
@@ -503,6 +534,7 @@ ItclCreateMemberFunc(
         Tcl_IncrRefCount(imPtr->origArgsPtr);
     }
     imPtr->codePtr    = mcode;
+    ItclPreserveMemberCode(mcode);
 
     if (imPtr->protection == ITCL_DEFAULT_PROTECT) {
         imPtr->protection = ITCL_PUBLIC;
@@ -621,22 +653,31 @@ ItclCreateMemberFunc(
             imPtr->flags |= ITCL_COMMON;
 	}
     }
-    if (strcmp(name, "___constructor_init") == 0) {
-        imPtr->flags |= ITCL_CONINIT;
-        iclsPtr->constructorInit = imPtr;
-    }
     if (strcmp(name, "constructor") == 0) {
+	/*
+	 * REVISE mcode->bodyPtr here!
+	 * Include a [my ItclConstructBase $iclsPtr] method call.
+	 * Inherited from itcl::Root
+	 */
+
+	Tcl_Obj *newBody = Tcl_NewStringObj("", -1);
+	Tcl_AppendToObj(newBody, 
+		"[::info object namespace ${this}]::my ItclConstructBase ", -1);
+	Tcl_AppendObjToObj(newBody, iclsPtr->fullNamePtr);
+	Tcl_AppendToObj(newBody, "\n", -1);
+
+	Tcl_AppendObjToObj(newBody, mcode->bodyPtr);
+	Tcl_DecrRefCount(mcode->bodyPtr);
+	mcode->bodyPtr = newBody;
+	Tcl_IncrRefCount(mcode->bodyPtr);
         imPtr->flags |= ITCL_CONSTRUCTOR;
-        iclsPtr->constructor = imPtr;
     }
     if (strcmp(name, "destructor") == 0) {
         imPtr->flags |= ITCL_DESTRUCTOR;
-        iclsPtr->destructor = imPtr;
     }
 
     Tcl_SetHashValue(hPtr, (ClientData)imPtr);
-    Itcl_PreserveData((ClientData)imPtr);
-    Itcl_EventuallyFree((ClientData)imPtr, Itcl_DeleteMemberFunc);
+    imPtr->refCount = 1;
 
     *imPtrPtr = imPtr;
     return TCL_OK;
@@ -730,13 +771,30 @@ Itcl_ChangeMemberFunc(
         return TCL_ERROR;
     }
 
+    if (imPtr->flags & ITCL_CONSTRUCTOR) {
+	/*
+	 * REVISE mcode->bodyPtr here!
+	 * Include a [my ItclConstructBase $iclsPtr] method call.
+	 * Inherited from itcl::Root
+	 */
+
+	Tcl_Obj *newBody = Tcl_NewStringObj("", -1);
+	Tcl_AppendToObj(newBody, 
+		"[::info object namespace ${this}]::my ItclConstructBase ", -1);
+	Tcl_AppendObjToObj(newBody, imPtr->iclsPtr->fullNamePtr);
+	Tcl_AppendToObj(newBody, "\n", -1);
+
+	Tcl_AppendObjToObj(newBody, mcode->bodyPtr);
+	Tcl_DecrRefCount(mcode->bodyPtr);
+	mcode->bodyPtr = newBody;
+	Tcl_IncrRefCount(mcode->bodyPtr);
+    }
+
     /*
      *  Free up the old implementation and install the new one.
      */
-    Itcl_PreserveData((ClientData)mcode);
-    Itcl_EventuallyFree((ClientData)mcode, Itcl_DeleteMemberCode);
-
-    Itcl_ReleaseData((ClientData)imPtr->codePtr);
+    ItclPreserveMemberCode(mcode);
+    ItclReleaseMemberCode(imPtr->codePtr);
     imPtr->codePtr = mcode;
     if (mcode->flags & ITCL_IMPLEMENT_TCL) {
 	ClientData pmPtr;
@@ -748,7 +806,6 @@ Itcl_ChangeMemberFunc(
                 (char *)imPtr->tmPtr, &isNewEntry);
         if (isNewEntry) {
             Tcl_SetHashValue(hPtr, imPtr);
-	    Itcl_PreserveData(imPtr->iclsPtr);
         }
     }
     ItclAddClassFunctionDictInfo(interp, imPtr->iclsPtr, imPtr);
@@ -894,9 +951,6 @@ ItclCreateMemberCode(
 	    if (strcmp(body, "@itcl-builtin-configure") == 0) {
 	        isDone = 1;
 	    }
-	    if (strcmp(body, "@itcl-builtin-info") == 0) {
-	        isDone = 1;
-	    }
 	    if (strcmp(body, "@itcl-builtin-isa") == 0) {
 	        isDone = 1;
 	    }
@@ -976,6 +1030,18 @@ ItclCreateMemberCode(
                     return TCL_ERROR;
                 }
 
+	/*
+	 * WARNING! WARNING! WARNING!
+	 * This is a pretty dangerous approach.  What's done here is
+	 * to copy over the proc + clientData implementation that
+	 * happens to be in place at the moment the method is
+	 * (re-)defined.  This denies any freedom for the clientData
+	 * to be changed dynamically or for the implementation to
+	 * shift from OBJCMD to ARGCMD or vice versa, which the
+	 * Itcl_Register(Obj)C routines explicitly permit.  The whole
+	 * system also lacks any scheme to unregister.
+	 */
+
                 if (objCmdProc != NULL) {
                     mcode->flags |= ITCL_IMPLEMENT_OBJCMD;
                     mcode->cfunc.objCmd = objCmdProc;
@@ -1000,6 +1066,7 @@ ItclCreateMemberCode(
     }
 
     *mcodePtr = mcode;
+
     return TCL_OK;
 }
 
@@ -1012,15 +1079,13 @@ ItclCreateMemberCode(
  *  of the function.  If the body is of the form "@name", then it is
  *  treated as a label for a C procedure registered by Itcl_RegisterC().
  *
- *  The implementation is kept by the member function definition, and
- *  controlled by a preserve/release paradigm.  That way, if it is in
- *  use while it is being redefined, it will stay around long enough
- *  to avoid a core dump.
+ *  A member function definition holds a handle for the implementation, and
+ *  calls ItclReleaseMemberCode when finished with it.
  *
  *  If any errors are encountered, this procedure returns TCL_ERROR
  *  along with an error message in the interpreter.  Otherwise, it
- *  returns TCL_OK, and "mcodePtr" returns a pointer to the new
- *  implementation.
+ *  returns TCL_OK, and stores a pointer to the new implementation in
+ *  "mcodePtr".
  * ------------------------------------------------------------------------
  */
 int
@@ -1040,17 +1105,13 @@ Itcl_CreateMemberCode(
  *  Itcl_DeleteMemberCode()
  *
  *  Destroys all data associated with the given command implementation.
- *  Invoked automatically by Itcl_ReleaseData() when the implementation
+ *  Invoked automatically by ItclReleaseData() when the implementation
  *  is no longer being used.
  * ------------------------------------------------------------------------
  */
-void
-Itcl_DeleteMemberCode(
-    char* cdata)  /* pointer to member code definition */
+void ItclFreeMemberCode (
+    ItclMemberCode* mCodePtr)
 {
-    ItclMemberCode* mCodePtr;
-    
-    mCodePtr = (ItclMemberCode*)cdata;
     if (mCodePtr == NULL) {
         return;
     }
@@ -1070,6 +1131,14 @@ Itcl_DeleteMemberCode(
     ckfree((char*)mCodePtr);
 }
 
+
+void
+Itcl_DeleteMemberCode(
+    char* cdata)  /* pointer to member code definition */
+{
+    ItclReleaseMemberCode((ItclMemberCode *)cdata);
+}
+
 
 /*
  * ------------------------------------------------------------------------
@@ -1193,21 +1262,6 @@ CallItclObjectCmd(
  *  result string or an error message in the interpreter.
  * ------------------------------------------------------------------------
  */
-static int
-CallConstructBase(
-    ClientData data[],
-    Tcl_Interp *interp,
-    int result)
-{
-    ItclMemberFunc *imPtr = data[0];
-    ItclObject *contextIoPtr = data[1];
-    int objc = PTR2INT(data[2]);
-    Tcl_Obj *const* objv = data[3];
-
-    result = Itcl_ConstructBase(interp, contextIoPtr, imPtr->iclsPtr,
-	        objc, objv);
-    return result;
-}
 
 int
 Itcl_EvalMemberCode(
@@ -1237,31 +1291,8 @@ Itcl_EvalMemberCode(
      *  Bump the reference count on this code, in case it is
      *  redefined or deleted during execution.
      */
-    Itcl_PreserveData((ClientData)mcode);
+    ItclPreserveMemberCode(mcode);
 
-    /*
-     *  If this code is a constructor, and if it is being invoked
-     *  when an object is first constructed (i.e., the "constructed"
-     *  table is still active within the object), then handle the
-     *  "initCode" associated with the constructor and make sure that
-     *  all base classes are properly constructed.
-     *
-     *  TRICKY NOTE:
-     *    The "initCode" must be executed here.  This is the only
-     *    opportunity where the arguments of the constructor are
-     *    available in a call frame.
-     */
-    if ((imPtr->flags & ITCL_CONSTRUCTOR) && (contextIoPtr != NULL) &&
-        contextIoPtr->constructed) {
-
-        callbackPtr = Itcl_GetCurrentCallbackPtr(interp);
-        Tcl_NRAddCallback(interp, CallConstructBase, imPtr, contextIoPtr,
-	        INT2PTR(objc), (ClientData)objv);
-        result = Itcl_NRRunCallbacks(interp, callbackPtr);
-        if (result != TCL_OK) {
-            goto evalMemberCodeDone;
-        }
-    }
     if ((imPtr->flags & ITCL_DESTRUCTOR) && (contextIoPtr != NULL)) {
         contextIoPtr->destructorHasBeenCalled = 1;
     }
@@ -1271,7 +1302,6 @@ Itcl_EvalMemberCode(
      */
     if (((mcode->flags & ITCL_IMPLEMENT_OBJCMD) != 0) ||
             ((mcode->flags & ITCL_IMPLEMENT_ARGCMD) != 0)) {
-	Itcl_SetCallFrameNamespace(interp, imPtr->iclsPtr->nsPtr);
 
         if ((mcode->flags & ITCL_IMPLEMENT_OBJCMD) != 0) {
             result = (*mcode->cfunc.objCmd)(mcode->clientData,
@@ -1299,8 +1329,7 @@ Itcl_EvalMemberCode(
          }
     }
 
-evalMemberCodeDone:
-    Itcl_ReleaseData((ClientData)mcode);
+    ItclReleaseMemberCode(mcode);
     return result;
 }
 
@@ -1397,68 +1426,124 @@ EquivArgLists(
  *  is not active.
  * ------------------------------------------------------------------------
  */
+
+void
+Itcl_SetContext(
+    Tcl_Interp *interp,
+    ItclObject *ioPtr)
+{
+    int new;
+    Itcl_Stack *stackPtr;
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+            ITCL_INTERP_DATA, NULL);
+    Tcl_HashEntry *hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr, &new);
+    ItclCallContext *contextPtr
+	    = (ItclCallContext *) ckalloc(sizeof(ItclCallContext));
+
+    memset(contextPtr, 0, sizeof(ItclCallContext));
+    contextPtr->ioPtr = ioPtr;
+    contextPtr->refCount = 1;
+
+    if (!new) {
+	Tcl_Panic("frame already has context?!");
+    }
+    
+    stackPtr = (Itcl_Stack *) ckalloc(sizeof(Itcl_Stack));
+    Itcl_InitStack(stackPtr);
+    Tcl_SetHashValue(hPtr, stackPtr);
+
+    Itcl_PushStack(contextPtr, stackPtr);
+}
+
+void
+Itcl_UnsetContext(
+    Tcl_Interp *interp)
+{
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+            ITCL_INTERP_DATA, NULL);
+    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr);
+    Itcl_Stack *stackPtr = (Itcl_Stack *) Tcl_GetHashValue(hPtr);
+    ItclCallContext *contextPtr = Itcl_PopStack(stackPtr);
+
+    if (Itcl_GetStackSize(stackPtr) > 0) {
+	Tcl_Panic("frame context stack not empty!");
+    }
+    Itcl_DeleteStack(stackPtr);
+    ckfree((char *) stackPtr);
+    Tcl_DeleteHashEntry(hPtr);
+    if (--contextPtr->refCount) {
+	Tcl_Panic("frame context ref count not zero!");
+    }
+    ckfree((char *)contextPtr);
+}
+
 int
 Itcl_GetContext(
     Tcl_Interp *interp,           /* current interpreter */
     ItclClass **iclsPtrPtr,       /* returns:  class definition or NULL */
     ItclObject **ioPtrPtr)        /* returns:  object data or NULL */
 {
-    Tcl_Namespace *activeNs = Tcl_GetCurrentNamespace(interp);
-    Tcl_HashEntry *hPtr;
-    ItclCallContext *callContextPtr;
-    ItclObjectInfo *infoPtr;
+    Tcl_Namespace *nsPtr;
 
-    /*
-     *  Return NULL for anything that cannot be found.
-     */
-    *ioPtrPtr = NULL;
+    /* Fetch the current call frame.  That determines context. */
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
 
-    if (!Itcl_IsClassNamespace(activeNs)) {
-        /*
-         *  If there is no class/object context, return an error message.
-         */
-        Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-            "namespace \"", activeNs->fullName, "\" is not a class namespace",
-            (char*)NULL);
-
-        return TCL_ERROR;
-    }
-    /*
-     *  If the active namespace is a class namespace, then return
-     *  all known info.  See if the current call frame is a known
-     *  object context, and if so, return that context.
-     */
-    infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
+    /* Try to map it to a context stack. */
+    ItclObjectInfo *infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
             ITCL_INTERP_DATA, NULL);
-    callContextPtr = Itcl_PeekStack(&infoPtr->contextStack);
-    if ((callContextPtr != NULL) && (callContextPtr->imPtr != NULL)) {
-        *iclsPtrPtr = callContextPtr->imPtr->iclsPtr;
-    } else {
-        hPtr = Tcl_FindHashEntry(&infoPtr->namespaceClasses,
-                (char *)activeNs);
-        if (hPtr != NULL) {
-            *iclsPtrPtr = (ItclClass *)Tcl_GetHashValue(hPtr);
-        }
-    }
-    if (*iclsPtrPtr == NULL) {
-        Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-            "namespace \"", activeNs->fullName, "\" is not a class namespace",
-            (char*)NULL);
+    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr);
+    if (hPtr) {
+	/* Frame maps to a context stack. */
+	Itcl_Stack *stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+	ItclCallContext *contextPtr = Itcl_PeekStack(stackPtr);
 
-        return TCL_ERROR;
-    }
+	assert(contextPtr);
 
-    if (callContextPtr == NULL) {
-	/* must be a class namespace without an object */
-	*ioPtrPtr = NULL;
+	if (contextPtr->objectFlags & ITCL_OBJECT_ROOT_METHOD) {
+	    ItclObject *ioPtr = contextPtr->ioPtr;
+
+	    *iclsPtrPtr = ioPtr->iclsPtr;
+	    *ioPtrPtr = ioPtr;
+	    return TCL_OK;
+	}
+
+	*iclsPtrPtr = contextPtr->imPtr ? contextPtr->imPtr->iclsPtr
+		: contextPtr->ioPtr->iclsPtr;
+	*ioPtrPtr = contextPtr->ioPtr ? contextPtr->ioPtr : infoPtr->currIoPtr;
 	return TCL_OK;
     }
-    *ioPtrPtr = callContextPtr->ioPtr;
-    if ((*ioPtrPtr == NULL) && ((*iclsPtrPtr)->nsPtr != NULL)) {
-        /* maybe we are in a constructor try currIoPtr */
-        *ioPtrPtr = (*iclsPtrPtr)->infoPtr->currIoPtr;
+
+    /* Frame has no Itcl context data.  No way to get object context. */
+    *ioPtrPtr = NULL;
+
+    /* Fall back to namespace for possible class context info. */
+    nsPtr = Tcl_GetCurrentNamespace(interp);
+    hPtr = Tcl_FindHashEntry(&infoPtr->namespaceClasses, (char *)nsPtr);
+    if (hPtr) {
+	*iclsPtrPtr = (ItclClass *)Tcl_GetHashValue(hPtr);
+
+	/*
+	 * DANGER! Following stanza of code was added to address a
+	 * regression from Itcl 4.0 -> Itcl 4.1 reported in Ticket
+	 * [c949e73d3e] without really understanding. May be trouble here!
+	 */
+	if ((*iclsPtrPtr)->nsPtr) {
+	    *ioPtrPtr = (*iclsPtrPtr)->infoPtr->currIoPtr;
+	}
+	return TCL_OK;
     }
-    return TCL_OK;
+
+    /* Cannot get any context.  Record an error message. */
+    if (interp) {
+        Tcl_SetObjResult(interp, Tcl_ObjPrintf(
+            "namespace \"%s\" is not a class namespace", nsPtr->fullName));
+    }
+    return TCL_ERROR;
 }
 
 /*
@@ -1645,9 +1730,9 @@ NRExecMethod(
      *  Execute the code for the method.  Be careful to protect
      *  the method in case it gets deleted during execution.
      */
-    Itcl_PreserveData((ClientData)imPtr);
+    ItclPreserveIMF(imPtr);
     result = Itcl_EvalMemberCode(interp, imPtr, ioPtr, objc, objv);
-    Itcl_ReleaseData((ClientData)imPtr);
+    ItclReleaseIMF(imPtr);
     return result;
 }
 
@@ -1731,11 +1816,11 @@ NRExecProc(
      *  Execute the code for the proc.  Be careful to protect
      *  the proc in case it gets deleted during execution.
      */
-    Itcl_PreserveData((ClientData)imPtr);
+    ItclPreserveIMF(imPtr);
 
     result = Itcl_EvalMemberCode(interp, imPtr, (ItclObject*)NULL,
         objc, objv);
-    Itcl_ReleaseData((ClientData)imPtr);
+    ItclReleaseIMF(imPtr);
     return result;
 }
 
@@ -1769,24 +1854,6 @@ CallInvokeMethodIfExists(
     }
     return TCL_OK;
 }
-
-static int
-CallPublicObjectCmd(
-    ClientData data[],
-    Tcl_Interp *interp,
-    int result)
-{
-    ItclClass *contextClass = data[0];
-    int cmdlinec = PTR2INT(data[1]);
-    Tcl_Obj **cmdlinev = data[2];
-
-    result = Itcl_PublicObjectCmd(contextClass->infoPtr->currIoPtr->oPtr,
-            interp, contextClass->clsPtr, cmdlinec, cmdlinev);
-    if (result != TCL_OK) {
-        return TCL_ERROR;
-    }
-    return TCL_OK;
-}
 /*
  * ------------------------------------------------------------------------
  *  Itcl_ConstructBase()
@@ -1805,58 +1872,25 @@ CallPublicObjectCmd(
  *  returns TCL_ERROR, along with an error message in the interpreter.
  * ------------------------------------------------------------------------
  */
+
 int
 Itcl_ConstructBase(
     Tcl_Interp *interp,       /* interpreter */
     ItclObject *contextObj,   /* object being constructed */
-    ItclClass *contextClass,  /* current class being constructed */
-    int objc,
-    Tcl_Obj *const *objv)
+    ItclClass *contextClass)  /* current class being constructed */
 {
-    int result;
+    int result = TCL_OK;
+    Tcl_Obj *objPtr;
     Itcl_ListElem *elem;
-    ItclClass *iclsPtr;
-    Tcl_HashEntry *entry;
-    Tcl_Obj *cmdlinePtr;
-    Tcl_Obj **cmdlinev;
-    void *callbackPtr;
-    int cmdlinec;
 
-    ItclShowArgs(1, "Itcl_ConstructBase", objc, objv);
     /*
      *  If the class has an "initCode", invoke it in the current context.
-     *
-     *  TRICKY NOTE:
-     *    This context is the call frame containing the arguments
-     *    for the constructor.  The "initCode" makes sense right
-     *    now--just before the body of the constructor is executed.
      */
-    Itcl_PushStack(contextClass, &contextClass->infoPtr->constructorStack);
-    if (contextClass->initCode) {
-        /*
-         *  Prepend the method name to the list of arguments.
-         */
-	int incr = 0;
-	if (strcmp(Tcl_GetString(objv[0]), "my") == 0) {
-	     /* number of args to skip depends on if we are called from
-	      * another constructor or directly */
-	     incr = 1;
-	}
-        cmdlinePtr = Itcl_CreateArgs(interp, "___constructor_init",
-	        objc-1-incr, objv+1+incr);
 
-        (void) Tcl_ListObjGetElements((Tcl_Interp*)NULL, cmdlinePtr,
-            &cmdlinec, &cmdlinev);
-        callbackPtr = Itcl_GetCurrentCallbackPtr(interp);
-        Tcl_NRAddCallback(interp, CallPublicObjectCmd, contextClass,
-	        INT2PTR(cmdlinec), cmdlinev, NULL);
-        result = Itcl_NRRunCallbacks(interp, callbackPtr);
-	Tcl_DecrRefCount(cmdlinev[0]);
-	Tcl_DecrRefCount(cmdlinev[1]);
-	Tcl_DecrRefCount(cmdlinePtr);
-        if (result != TCL_OK) {
-	    return result;
-	}
+    if (contextClass->initCode) {
+
+	/* TODO: NRE */
+	result = Tcl_EvalObj(interp, contextClass->initCode);
     }
 
     /*
@@ -1866,46 +1900,64 @@ Itcl_ConstructBase(
      *  in reverse order, so that least-specific classes are constructed
      *  first.
      */
-    elem = Itcl_LastListElem(&contextClass->bases);
-    while (elem != NULL) {
-        iclsPtr = (ItclClass*)Itcl_GetListValue(elem);
+
+    objPtr = Tcl_NewStringObj("constructor", -1);
+    Tcl_IncrRefCount(objPtr);
+    for (elem = Itcl_LastListElem(&contextClass->bases);
+	    result == TCL_OK && elem != NULL;
+	    elem = Itcl_PrevListElem(elem)) {
+	    
+	Tcl_HashEntry *entry;
+        ItclClass *iclsPtr = (ItclClass*)Itcl_GetListValue(elem);
 
         if (Tcl_FindHashEntry(contextObj->constructed,
-	        (char *)iclsPtr->namePtr) == NULL) {
-	    Tcl_Obj *objPtr;
+		(char *)iclsPtr->namePtr)) {
 
-            callbackPtr = Itcl_GetCurrentCallbackPtr(interp);
+	    /* Already constructed, nothing to do. */
+	    continue;
+	}
+
+        entry = Tcl_FindHashEntry(&iclsPtr->functions, (char *)objPtr);
+        if (entry) {
+            void *callbackPtr = Itcl_GetCurrentCallbackPtr(interp);
             Tcl_NRAddCallback(interp, CallInvokeMethodIfExists, iclsPtr,
 	            contextObj, INT2PTR(0), NULL);
             result = Itcl_NRRunCallbacks(interp, callbackPtr);
-            if (result != TCL_OK) {
-                return TCL_ERROR;
-            }
-            /*
-             *  The base class may not have a constructor, but its
-             *  own base classes could have one.  If the constructor
-             *  wasn't found in the last step, then other base classes
-             *  weren't constructed either.  Make sure that all of its
-             *  base classes are properly constructed.
-             */
-	    objPtr = Tcl_NewStringObj("constructor", -1);
-	    Tcl_IncrRefCount(objPtr);
-            entry = Tcl_FindHashEntry(&iclsPtr->functions, (char *)objPtr);
-	    Tcl_DecrRefCount(objPtr);
-            if (entry == NULL) {
-                result = Itcl_ConstructBase(interp, contextObj, iclsPtr,
-		        objc, objv);
-                if (result != TCL_OK) {
-                    return TCL_ERROR;
-                }
-            }
+	} else {
+            result = Itcl_ConstructBase(interp, contextObj, iclsPtr);
         }
-        elem = Itcl_PrevListElem(elem);
     }
-    Itcl_PopStack(&contextClass->infoPtr->constructorStack);
-    return TCL_OK;
+    Tcl_DecrRefCount(objPtr);
+    return result;
 }
 
+int
+ItclConstructGuts(
+    ItclObject *contextObj,
+    Tcl_Interp *interp,
+    int objc,
+    Tcl_Obj *const objv[])
+{
+    ItclClass *contextClass;
+
+    /* Ignore syntax error */
+    if (objc != 3) {
+	return TCL_OK;
+    }
+
+    /* Object is fully constructed. This becomes no-op. */
+    if (contextObj->constructed == NULL) {
+	return TCL_OK;
+    }
+
+    contextClass = Itcl_FindClass(interp, Tcl_GetString(objv[2]), 0);
+    if (contextClass == NULL) {
+	return TCL_OK;
+    }
+
+
+    return Itcl_ConstructBase(interp, contextObj, contextClass);
+}
 
 /*
  * ------------------------------------------------------------------------
@@ -1961,7 +2013,7 @@ Itcl_InvokeMethodIfExists(
          *  Execute the code for the method.  Be careful to protect
          *  the method in case it gets deleted during execution.
          */
-        Itcl_PreserveData((ClientData)imPtr);
+	ItclPreserveIMF(imPtr);
 
 	if (contextObjectPtr->oPtr == NULL) {
             Tcl_DecrRefCount(cmdlinePtr);
@@ -1969,9 +2021,7 @@ Itcl_InvokeMethodIfExists(
 	}
         result = Itcl_EvalMemberCode(interp, imPtr, contextObjectPtr,
 	        cmdlinec, cmdlinev);
-	Tcl_DecrRefCount(cmdlinev[0]);
-	Tcl_DecrRefCount(cmdlinev[1]);
-        Itcl_ReleaseData((ClientData)imPtr);
+	ItclReleaseIMF(imPtr);
         Tcl_DecrRefCount(cmdlinePtr);
     } else {
         if (contextClassPtr->flags &
@@ -2087,12 +2137,6 @@ Itcl_CmdAliasProc(
     hPtr = Tcl_FindHashEntry(&iclsPtr->resolveCmds, (char *)objPtr);
     Tcl_DecrRefCount(objPtr);
     if (hPtr == NULL) {
-	if (strcmp(cmdName, "info") == 0) {
-	    return Tcl_FindCommand(interp, "::itcl::builtin::Info", NULL, 0);
-	}
-	if (strcmp(cmdName, "@itcl-builtin-info") == 0) {
-	    return Tcl_FindCommand(interp, "::itcl::builtin::Info", NULL, 0);
-	}
 	if (strcmp(cmdName, "@itcl-builtin-cget") == 0) {
 	    return Tcl_FindCommand(interp, "::itcl::builtin::cget", NULL, 0);
 	}
@@ -2161,17 +2205,10 @@ Itcl_CmdAliasProc(
 	if (strcmp(cmdName, "@itcl-builtin-classunknown") == 0) {
 	    return Tcl_FindCommand(interp, "::itcl::builtin::classunknown", NULL, 0);
 	}
-	if (*cmdName == '@') {
-	    return Tcl_FindCommand(interp,
-	            ITCL_NAMESPACE"::methodset::callCCommand", NULL, 0);
-	}
         return NULL;
     }
     clookup = (ItclCmdLookup *)Tcl_GetHashValue(hPtr);
     imPtr = clookup->imPtr;
-    if (strcmp(cmdName, "info") == 0) {
-        return Tcl_FindCommand(interp, "::itcl::builtin::Info", NULL, 0);
-    }
     return imPtr->accessCmd;
 }
 
@@ -2294,6 +2331,7 @@ ItclCheckCallMethod(
     Tcl_CallFrame *framePtr,
     int *isFinished)
 {
+    Itcl_Stack *stackPtr;
 
     Tcl_Object oPtr;
     ItclObject *ioPtr;
@@ -2308,13 +2346,12 @@ ItclCheckCallMethod(
     int cObjc;
     int min_allowed_args;
 
+    ItclObjectInfo *infoPtr;
+
     oPtr = NULL;
     hPtr = NULL;
     imPtr = (ItclMemberFunc *)clientData;
-    Itcl_PreserveData(imPtr);
-    if (imPtr->codePtr != NULL) {
-        Itcl_PreserveData(imPtr->codePtr);
-    }
+    ItclPreserveIMF(imPtr);
     if (imPtr->flags & ITCL_CONSTRUCTOR) {
         ioPtr = imPtr->iclsPtr->infoPtr->currIoPtr;
     } else {
@@ -2352,6 +2389,13 @@ ItclCheckCallMethod(
 	result = TCL_ERROR;
 	goto finishReturn;
     }
+  if (framePtr) {
+    /*
+     * This stanza is in place to seize control over usage error messages
+     * before TclOO examines the arguments and produces its own.  This
+     * gives Itcl stability in its error messages at the cost of inconsistency
+     * with Tcl's evolving conventions.
+     */
     cObjc = Itcl_GetCallFrameObjc(interp);
     cObjv = Itcl_GetCallFrameObjv(interp);
     min_allowed_args = cObjc-2;
@@ -2359,23 +2403,16 @@ ItclCheckCallMethod(
         min_allowed_args++;
     }
     if (min_allowed_args < imPtr->argcount) {
-	if (strcmp(Tcl_GetString(imPtr->namePtr), "info") == 0) {
-            Tcl_Obj *objPtr = Tcl_NewStringObj(
-	            "wrong # args: should be one of...\n", -1);
-            ItclGetInfoUsage(interp, objPtr, imPtr->iclsPtr->infoPtr);
-	    Tcl_SetObjResult(interp, objPtr);
-	} else {
-            Tcl_AppendResult(interp, "wrong # args: should be \"",
-	            Tcl_GetString(cObjv[0]), " ",
-	            Tcl_GetString(imPtr->namePtr), " ",
-		    Tcl_GetString(imPtr->usagePtr), "\"", NULL);
-	}
+	Tcl_AppendResult(interp, "wrong # args: should be \"",
+		Tcl_GetString(cObjv[0]), " ", Tcl_GetString(imPtr->namePtr),
+		" ", Tcl_GetString(imPtr->usagePtr), "\"", NULL);
         if (isFinished != NULL) {
             *isFinished = 1;
         }
 	result = TCL_ERROR;
 	goto finishReturn;
     }
+  }
     isNew = 0;
     callContextPtr = NULL;
     currNsPtr = Tcl_GetCurrentNamespace(interp);
@@ -2416,12 +2453,42 @@ ItclCheckCallMethod(
     if (isNew) {
         Tcl_SetHashValue(hPtr, callContextPtr);
     }
-    Itcl_PushStack(callContextPtr, &imPtr->iclsPtr->infoPtr->contextStack);
+
+    if (framePtr == NULL) {
+	framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+    }
+
+    isNew = 0;
+    infoPtr = imPtr->iclsPtr->infoPtr;
+    hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)framePtr, &isNew);
+    if (isNew) {
+	stackPtr = (Itcl_Stack *)ckalloc(sizeof(Itcl_Stack));
+	Itcl_InitStack(stackPtr);
+        Tcl_SetHashValue(hPtr, stackPtr);
+    } else {
+	stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    }
+
+    assert (callContextPtr) ;
+    Itcl_PushStack(callContextPtr, stackPtr);
+
+    /* Ugly abuse alert.  Two maps in one table */
+    hPtr = Tcl_CreateHashEntry(&infoPtr->frameContext,
+	    (char *)contextPtr, &isNew);
+    if (isNew) {
+	stackPtr = (Itcl_Stack *)ckalloc(sizeof(Itcl_Stack));
+	Itcl_InitStack(stackPtr);
+        Tcl_SetHashValue(hPtr, stackPtr);
+    } else {
+	stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    }
+
+    Itcl_PushStack(framePtr, stackPtr);
 
     if (ioPtr != NULL) {
         ioPtr->callRefCount++;
-	Itcl_PreserveData(ioPtr);
-	Itcl_PreserveData(ioPtr->iclsPtr);
+	ItclPreserveObject(ioPtr);
     }
     imPtr->iclsPtr->callRefCount++;
     if (!imPtr->iclsPtr->infoPtr->useOldResolvers) {
@@ -2434,10 +2501,7 @@ ItclCheckCallMethod(
     }
     return result;
 finishReturn:
-    Itcl_ReleaseData(imPtr);
-    if (imPtr->codePtr != NULL) {
-        Itcl_ReleaseData(imPtr->codePtr);
-    }
+    ItclReleaseIMF(imPtr);
     return result;
 }
 
@@ -2464,10 +2528,31 @@ ItclAfterCallMethod(
     int result;
 
     imPtr = (ItclMemberFunc *)clientData;
-
     callContextPtr = NULL;
     if (contextPtr != NULL) {
-        callContextPtr = Itcl_PopStack(&imPtr->infoPtr->contextStack);
+    ItclObjectInfo *infoPtr = imPtr->infoPtr;
+    Tcl_CallFrame *framePtr;
+    Itcl_Stack *stackPtr;
+	
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)contextPtr);
+    assert(hPtr);
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    framePtr = Itcl_PopStack(stackPtr);
+    if (Itcl_GetStackSize(stackPtr) == 0) {
+	Itcl_DeleteStack(stackPtr);
+	ckfree((char *) stackPtr);
+	Tcl_DeleteHashEntry(hPtr);
+    }
+
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)framePtr);
+    assert(hPtr);
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    callContextPtr = Itcl_PopStack(stackPtr);
+    if (Itcl_GetStackSize(stackPtr) == 0) {
+	Itcl_DeleteStack(stackPtr);
+	ckfree((char *) stackPtr);
+	Tcl_DeleteHashEntry(hPtr);
+    }
     }
     if (callContextPtr == NULL) {
         if ((imPtr->flags & ITCL_COMMON) ||
@@ -2489,6 +2574,7 @@ ItclAfterCallMethod(
      */
     ioPtr = callContextPtr->ioPtr;
     if (ioPtr != NULL) {
+      if (imPtr->iclsPtr) {
         imPtr->iclsPtr->callRefCount--;
         if (imPtr->flags & (ITCL_CONSTRUCTOR | ITCL_DESTRUCTOR)) {
             if ((imPtr->flags & ITCL_DESTRUCTOR) && ioPtr &&
@@ -2502,6 +2588,7 @@ ItclAfterCallMethod(
                     (char *)imPtr->iclsPtr->namePtr, &newEntry);
             }
         }
+      }
         ioPtr->callRefCount--;
         if (ioPtr->flags & ITCL_OBJECT_SHOULD_VARNS_DELETE) {
             ItclDeleteObjectVariablesNamespace(interp, ioPtr);
@@ -2516,18 +2603,14 @@ ItclAfterCallMethod(
             if (hPtr == NULL) {
                 ckfree((char *)callContextPtr);
 	    }
-            Itcl_ReleaseData(ioPtr->iclsPtr);
-            Itcl_ReleaseData(ioPtr);
+	    ItclReleaseObject(ioPtr);
         } else {
             ckfree((char *)callContextPtr);
         }
     }
     result = call_result;
 finishReturn:
-    if (imPtr->codePtr != NULL) {
-        Itcl_ReleaseData(imPtr->codePtr);
-    }
-    Itcl_ReleaseData(imPtr);
+    ItclReleaseIMF(imPtr);
     return result;
 }
 
@@ -2537,7 +2620,6 @@ ItclProcErrorProc(
     Tcl_Obj *procNameObj)
 {
     Tcl_Obj *objPtr;
-    Tcl_Namespace *upNsPtr;
     Tcl_HashEntry *hPtr;
     ItclObjectInfo *infoPtr;
     ItclCallContext *callContextPtr;
@@ -2545,52 +2627,37 @@ ItclProcErrorProc(
     ItclObject *contextIoPtr;
     ItclClass *currIclsPtr;
     char num[20];
-    int constructorStackIndex;
-    int constructorStackSize;
-    int isFirstLoop;
-    int loopCnt;
+    Itcl_Stack *stackPtr;
 
+    /* Fetch the current call frame.  That determines context. */
+    Tcl_CallFrame *framePtr = Itcl_GetUplevelCallFrame(interp, 0);
+
+    /* Try to map it to a context stack. */
     infoPtr = (ItclObjectInfo *)Tcl_GetAssocData(interp,
             ITCL_INTERP_DATA, NULL);
-    callContextPtr = Itcl_PeekStack(&infoPtr->contextStack);
+    hPtr = Tcl_FindHashEntry(&infoPtr->frameContext, (char *)framePtr);
+    if (hPtr == NULL) {
+	/* Can this happen? */
+	return;
+    }
+
+    /* Frame maps to a context stack. */
+    stackPtr = (Itcl_Stack *)Tcl_GetHashValue(hPtr);
+    callContextPtr = Itcl_PeekStack(stackPtr);
+
+    if (callContextPtr == NULL) {
+	return;
+    }
+
     currIclsPtr = NULL;
-    loopCnt = 1;
-    isFirstLoop = 1;
-    upNsPtr = Itcl_GetUplevelNamespace(interp, 1);
-    constructorStackIndex = -1;
     objPtr = NULL;
-    while ((callContextPtr != NULL) && (loopCnt > 0)) {
+    {
 	imPtr = callContextPtr->imPtr;
         contextIoPtr = callContextPtr->ioPtr;
         objPtr = Tcl_NewStringObj("\n    ", -1);
 
         if (imPtr->flags & ITCL_CONSTRUCTOR) {
-	    /* have to look for classes in construction where the constructor
-	     * has not yet been called, but only the initCode or the
-	     * inherited constructors
-	     */
-            if (isFirstLoop) {
-	        isFirstLoop = 0;
-                constructorStackSize = Itcl_GetStackSize(
-		        &imPtr->iclsPtr->infoPtr->constructorStack);
-	        constructorStackIndex = constructorStackSize;
-	        currIclsPtr = imPtr->iclsPtr;
-	    } else {
-	        currIclsPtr = (ItclClass *)Itcl_GetStackValue(
-	                &imPtr->iclsPtr->infoPtr->constructorStack,
-		        constructorStackIndex);
-            }
-	    if (constructorStackIndex < 0) {
-	        break;
-	    }
-	    if (currIclsPtr == NULL) {
-	        break;
-	    }
-	    if (upNsPtr == currIclsPtr->nsPtr) {
-	        break;
-	    }
-	    constructorStackIndex--;
-	    loopCnt++;
+	    currIclsPtr = imPtr->iclsPtr;
             Tcl_AppendToObj(objPtr, "while constructing object \"", -1);
             Tcl_GetCommandFullName(interp, contextIoPtr->accessCmd, objPtr);
             Tcl_AppendToObj(objPtr, "\" in ", -1);
@@ -2600,17 +2667,6 @@ ItclProcErrorProc(
                 Tcl_AppendToObj(objPtr, " (", -1);
             }
         }
-        if (imPtr->flags & ITCL_CONINIT) {
-            Tcl_AppendToObj(objPtr, "while constructing object \"", -1);
-            Tcl_GetCommandFullName(interp, contextIoPtr->accessCmd, objPtr);
-            Tcl_AppendToObj(objPtr, "\" in ", -1);
-            Tcl_AppendToObj(objPtr,
-	            Tcl_GetString(imPtr->iclsPtr->fullNamePtr), -1);
-            Tcl_AppendToObj(objPtr, "::constructor", -1);
-            if ((imPtr->codePtr->flags & ITCL_IMPLEMENT_TCL) != 0) {
-                Tcl_AppendToObj(objPtr, " (", -1);
-            }
-        }
 	if (imPtr->flags & ITCL_DESTRUCTOR) {
 	    contextIoPtr->flags = 0;
 	    Tcl_AppendToObj(objPtr, "while deleting object \"", -1);
@@ -2621,7 +2677,7 @@ ItclProcErrorProc(
                 Tcl_AppendToObj(objPtr, " (", -1);
             }
         }
-	if (!(imPtr->flags & (ITCL_CONSTRUCTOR|ITCL_DESTRUCTOR|ITCL_CONINIT))) {
+	if (!(imPtr->flags & (ITCL_CONSTRUCTOR|ITCL_DESTRUCTOR))) {
             Tcl_AppendToObj(objPtr, "(", -1);
 
 	    hPtr = Tcl_FindHashEntry(&infoPtr->objects, (char *)contextIoPtr);
@@ -2682,7 +2738,6 @@ ItclProcErrorProc(
 
         Tcl_AppendObjToErrorInfo(interp, objPtr);
 	objPtr = NULL;
-        loopCnt--;
     }
     if (objPtr != NULL) {
         Tcl_DecrRefCount(objPtr);
